"use strict";
var fs_1 = require('fs');
var path_1 = require('path');
var source_map_1 = require('source-map');
var html_minifier_1 = require('html-minifier');
var module_exists_1 = require('./module-exists');
function mahaloTransformer(fileName, extension, minimize) {
    if (minimize === void 0) { minimize = true; }
    var text = fs_1.readFileSync(fileName).toString();
    var map = fileName + '.ts';
    var context = path_1.dirname(fileName);
    var components = [];
    var behaviors = [];
    var styles = [];
    var images = [];
    var mapGenerator = new source_map_1.SourceMapGenerator();
    mapGenerator.addMapping({
        generated: { line: 1, column: 0 },
        original: { line: 1, column: 0 },
        source: path_1.basename(fileName)
    });
    extension = '.' + (extension || 'mhml');
    // Add dummy file for nice IDE inegration
    fs_1.existsSync(map) || fs_1.writeFileSync(map, 'var Template: ITemplate;\n\nexport default Template;');
    minimize && (text = html_minifier_1.minify(text, {
        removeComments: true,
        removeCommentsFromCDATA: true,
        removeCDATASectionsFromCDATA: true,
        collapseWhitespace: true,
        conservativeCollapse: true,
        keepClosingSlash: true,
        minifyJS: true,
        minifyCSS: true,
        removeScriptTypeAttributes: true,
        removeRedundantAttributes: true
    }));
    // Find use tags and handle them
    text = text.replace(USE_TAG, function (m, kind, path, _, selector) {
        resolveModule(kind, path, selector, components, behaviors);
        return '';
    });
    // Find stylesheets and handle them
    text = text.replace(LINK_TAG, function (m, _, path) {
        styles.push("require('" + path + "')");
        return '';
    });
    // Find images and handle them (step 1)
    text = text.replace(IMG_TAG, function (m, _, path, __) {
        var i = images.length;
        path = /^\//.test(path) ? process.cwd().replace(/\\/g, '/') + path : path;
        images.push("\" + require('" + path + "') + \"");
        return '<img ' + _ + 'src="###require' + i + '"' + __ + '>';
    });
    // Make sure content is an exportable string
    text = JSON.stringify(text);
    // Find images and handle them (step 2)
    text = text.replace(/###require(\d+)/g, function (m, i) { return images[i]; });
    return {
        fileName: fileName,
        text: compile(components, behaviors, styles, text),
        map: JSON.parse(mapGenerator.toString())
    };
    //////////
    function resolveModule(kind, path, selector, components, behaviors) {
        var desc = {
            selector: getSelector(selector, path, kind),
            path: path,
            files: []
        };
        if (kind === 'component') {
            components.push(desc);
            module_exists_1.default(context, path, '.ts') && addComponent(desc);
            module_exists_1.default(context, path, extension) && addTemplate(desc, extension);
        }
        else {
            behaviors.push(desc);
            module_exists_1.default(context, path, '.ts') && addBehavior(desc);
        }
    }
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = mahaloTransformer;
//////////
var USE_TAG = /<use\s*(component|behavior)="(~?[\w\/.-]+)"(\s*select="([^"]+)")?\s*\/>\s*/ig, LINK_TAG = /<link\s(.*?)href="(~?[\w\/.-]*)"[^>]*>\s*/ig, IMG_TAG = /<img\s(.*?)src="(~?[\w\/.-]*)"([^>]*)>/ig, VALID_SELECTOR = /^((\.-?)?[\w][\w-]*|\[[^\s"'>\/=]+([*^]?=(["'])[^\4]*\4)?](?![\w]))+$/, VALID_ATTRIBUTE = /^[^\s"'>\/=]+$/;
function addComponent(desc) {
    desc.files.push("Component: require('" + desc.path + "')['default']");
}
function addTemplate(desc, extension) {
    desc.files.push("template: require('" + desc.path + extension + "')['default']");
}
function addBehavior(desc) {
    desc.files.push("'" + desc.selector + "': require('" + desc.path + "')['default']");
}
function getSelector(selector, path, kind) {
    if (selector) {
        if (kind === 'component') {
            if (!VALID_SELECTOR.test(selector)) {
                throw Error('Invalid selector for component');
            }
        }
        else if (!VALID_ATTRIBUTE.test(selector)) {
            throw Error('Invalid attribute name for behavior.');
        }
    }
    else {
        selector = path.split('/').pop().replace(/[^\w]/g, '-').replace(/^-/, '');
    }
    return selector;
}
function compile(components, behaviors, styles, text) {
    var _components = [], _behaviors = [];
    components.forEach(function (component) {
        component.files.length && _components.push("'" + component.selector + "': {\n\t\t\t" + component.files.join(',\n\t\t\t') + "\n\t\t}");
    });
    behaviors.forEach(function (behavior) {
        behavior.files.length && _behaviors.push(behavior.files[0]);
    });
    return "\"use strict\";\n\nObject.defineProperty(exports, \"__esModule\", {\n    value: true\n});\n" + (styles.length ? '\n' + styles.join(';\n') + ';\n' : '') + "\nvar Template = require('mahalo')['Template'],\n    components = {\n\t\t" + _components.join(',\n\t\t') + "\n\t},\n    behaviors = {\n\t\t" + _behaviors.join(',\n\t\t') + "\n\t};\t\t\n\nexports.default = new Template(" + text + ", components, behaviors);";
}
